#!/usr/bin/env python3
# encoding: utf-8
#
# This file is part of ckanext-doi
# Created by the Natural History Museum in London, UK

import datetime

PKG_DICT = {
    'domain': 'data.nhm.ac.uk',
    'license_title': 'CC0-1.0',
    'maintainer': 'Maintainer, Data',
    'relationships_as_object': [],
    'doi_status': False,
    'private': False,
    'maintainer_email': None,
    'num_tags': 2,
    'id': '919379ef-ea08-4a5f-b370-7845e24117b4',
    'metadata_created': '2020-11-09T17:14:06.700561',
    'owner_org': '2dc787a8-17c3-4425-bed7-2b97c81789d0',
    'metadata_modified': '2020-11-09T17:14:07.225364',
    'author': 'Author, Test',
    'author_email': None,
    'state': 'active',
    'version': '1',
    'license_id': 'CC0-1.0',
    'type': 'dataset',
    'resources': [],
    'num_resources': 0,
    'tags': [
        {
            'vocabulary_id': None,
            'state': 'active',
            'display_name': 'not-a-real-package',
            'id': 'a3169a9a-e03c-4844-8694-0bd0cfaaaeae',
            'name': 'not-a-real-package',
        },
        {
            'vocabulary_id': None,
            'state': 'active',
            'display_name': 'testing',
            'id': 'cbd1ad22-d75e-4641-8d47-d29bbd758e0c',
            'name': 'testing',
        },
    ],
    'groups': [],
    'creator_user_id': '21ff8eeb-38e5-4348-8c1d-f409b2a9ca03',
    'relationships_as_subject': [],
    'doi_publisher': 'Natural History Museum',
    'doi': '10.4124/gqcx00k4',
    'name': 'test_pkg',
    'isopen': False,
    'url': None,
    'notes': 'These are some notes about the package.',
    'title': 'Test Package',
    'extras': [],
    'organization': {
        'description': 'Just another test organization.',
        'title': 'Test Organization',
        'created': '2020-11-09T17:14:06.583496',
        'approval_status': 'approved',
        'is_organization': True,
        'state': 'active',
        'image_url': 'http://placekitten.com/g/200/100',
        'revision_id': '2b0d57a6-5d2f-46c3-859e-32635bfcf741',
        'type': 'organization',
        'id': '2dc787a8-17c3-4425-bed7-2b97c81789d0',
        'name': 'test_org_00',
    },
    'revision_id': 'c71800fb-60b7-421f-a7f2-56dc8efaf21d',
    'doi_date_published': None,
}

METADATA_DICT = {
    'fundingReferences': [],
    'contributors': [
        {'full_name': 'Author, Test', 'contributor_type': 'Researcher'},
        {'full_name': 'Maintainer, Data', 'contributor_type': 'DataManager'},
    ],
    'geolocations': [],
    'dates': [
        {
            'date': datetime.datetime(2020, 11, 9, 17, 14, 6, 700561),
            'dateType': 'Created',
        },
        {
            'date': datetime.datetime(2020, 11, 9, 17, 14, 7, 225364),
            'dateType': 'Updated',
        },
        {'date': None, 'dateType': 'Issued'},
    ],
    'relatedIdentifiers': [],
    'titles': [{'title': 'Test Package'}],
    'descriptions': [
        {
            'descriptionType': 'Other',
            'description': 'These are some notes about the package.',
        }
    ],
    'subjects': [],
    'publisher': 'Natural History Museum',
    'publicationYear': 2020,
    'rightsList': [],
    'language': '',
    'sizes': ['0 kb'],
    'resourceType': 'dataset',
    'alternateIdentifiers': [
        {
            'alternateIdentifierType': 'URL',
            'alternateIdentifier': 'http://data.nhm.ac.uk/dataset/919379ef-ea08-4a5f-b370-7845e24117b4',
        }
    ],
    'version': '1',
    'formats': [],
    'creators': [{'full_name': 'Author, Test'}],
}

# minimal example from https://datacite.readthedocs.io/en/latest
XML_DICT = {
    'identifiers': [
        {
            'identifierType': 'DOI',
            'identifier': '10.4124/abcd1234',
        }
    ],
    'creators': [
        {'name': 'Smith, John'},
    ],
    'titles': [
        {
            'title': 'Minimal Test Case',
        }
    ],
    'publisher': 'Invenio Software',
    'publicationYear': '2015',
    'types': {'resourceType': 'Dataset', 'resourceTypeGeneral': 'Dataset'},
    'schemaVersion': 'http://datacite.org/schema/kernel-4',
}



#!/usr/bin/env python3
# encoding: utf-8
#
# This file is part of ckanext-doi
# Created by the Natural History Museum in London, UK


def create_contributor(
    full_name=None,
    family_name=None,
    given_name=None,
    is_org=False,
    contributor_type=None,
    affiliations=None,
    identifiers=None,
):
    """
    Create a dictionary representation of a contributing entity (either a person or an
    organisation) for use in an xml_dict.

    :param full_name: the full name of the creator, in the format "FamilyName, GivenName";
    can be omitted if family_name and given_name are provided
    :param family_name: family name of the creator; will be ignored if given_name is None
    :param given_name: given name(s) or initials of the creator; will be ignored if
    family_name is None
    :param is_org: sets name type to Organizational if true
    :param contributor_type: the contributor type to set
    :param affiliations: affiliations of the contributor, either a string or list of strings
    :param identifiers: a list of dicts with "identifier", "scheme", and (optionally) "scheme_uri"
    :return: a dict
    """
    if is_org and full_name is None:
        raise ValueError('Creator name must be supplied as full_name="Org Name"')
    if full_name is None and (family_name is None or given_name is None):
        raise ValueError(
            'Creator name must be supplied, either as full_name="FamilyName, '
            'GivenName" or separately as family_name and given_name'
        )
    if full_name is None:
        full_name = f'{family_name}, {given_name}'
    if (family_name is None or given_name is None) and not is_org:
        # try to extract the family and given names from the full name
        name_parts = full_name.split(',', 1)
        if len(name_parts) > 1:
            family_name, given_name = [x.strip() for x in name_parts]
        else:
            # assume it was formatted incorrectly and split by spaces instead
            # if that doesn't work then there isn't much we can do
            name_parts = full_name.split(' ')
            family_name = name_parts[-1].strip()
            given_name = ' '.join(name_parts[0:-1]).strip()
    contributor = {
        'name': full_name,
        'nameType': 'Organizational' if is_org else 'Personal',
    }
    if not is_org:
        contributor['familyName'] = family_name
        contributor['givenName'] = given_name
    if contributor_type is not None:
        contributor['contributorType'] = contributor_type
    if affiliations is not None:
        contributor['affiliations'] = []
        if isinstance(affiliations, str):
            affiliations = [affiliations]
        for affiliation in affiliations:
            contributor['affiliations'].append({'affiliation': affiliation})
    if identifiers:
        contributor['nameIdentifiers'] = []
        for _id in identifiers:
            if 'identifier' not in _id or 'scheme' not in _id:
                continue
            id_dict = {
                'nameIdentifier': _id['identifier'],
                'nameIdentifierScheme': _id['scheme'],
            }
            if 'scheme_uri' in _id:
                id_dict['schemeUri'] = _id['scheme_uri']
            contributor['nameIdentifiers'].append(id_dict)
    return contributor



#!/usr/bin/env python3
# encoding: utf-8
#
# This file is part of ckanext-doi
# Created by the Natural History Museum in London, UK

import logging

from ckan.lib.helpers import lang as ckan_lang
from ckan.model import Package
from ckan.plugins import PluginImplementations, toolkit

from ckanext.doi.interfaces import IDoi
from ckanext.doi.lib import xml_utils
from ckanext.doi.lib.errors import DOIMetadataException
from ckanext.doi.lib.helpers import date_or_none, get_site_url, package_get_year

log = logging.getLogger(__name__)


def build_metadata_dict(pkg_dict):
    """
    Build/extract a basic dict of metadata that can then be passed to build_xml_dict.

    :param pkg_dict: dict of package details
    """
    metadata_dict = {}

    # collect errors instead of throwing them immediately; some data may not be correctly handled
    # by this base method but will be handled correctly by plugins that implement IDoi
    errors = {}

    # required fields first (identifier will be added later)
    required = {
        'creators': [],
        'titles': [],
        'publisher': None,
        'publicationYear': None,
        'resourceType': None,
    }

    def _add_required(key, get_func):
        try:
            required[key] = get_func()
        except Exception as e:
            errors[key] = e

    # CREATORS
    _add_required('creators', lambda: [{'full_name': pkg_dict.get('author')}])

    # TITLES
    _add_required('titles', lambda: [{'title': pkg_dict.get('title')}])

    # PUBLISHER
    _add_required('publisher', lambda: toolkit.config.get('ckanext.doi.publisher'))

    # PUBLICATION YEAR
    _add_required('publicationYear', lambda: package_get_year(pkg_dict))

    # TYPE
    _add_required('resourceType', lambda: pkg_dict.get('type'))

    # now the optional fields
    optional = {
        'subjects': [],
        'contributors': [],
        'dates': [],
        'language': '',
        'alternateIdentifiers': [],
        'relatedIdentifiers': [],
        'sizes': [],
        'formats': [],
        'version': '',
        'rightsList': [],
        'descriptions': [],
        'geolocations': [],
        'fundingReferences': [],
    }

    # SUBJECTS
    # use the tag list
    try:
        tags = pkg_dict.get('tag_string', '').split(',')
        tags += [
            tag['name'] if isinstance(tag, dict) else tag
            for tag in pkg_dict.get('tags', [])
        ]
        optional['subjects'] = [
            {'subject': tag} for tag in sorted({t for t in tags if t != ''})
        ]
    except Exception as e:
        errors['subjects'] = e

    # CONTRIBUTORS
    # use the author and maintainer; no splitting or parsing for either
    # no try/except for this because it's just a simple .get() and if that doesn't work then we
    # want to know
    author = pkg_dict.get('author')
    maintainer = pkg_dict.get('maintainer')
    if author is not None:
        optional['contributors'].append(
            {'contributor_type': 'Researcher', 'full_name': author}
        )
    if maintainer is not None:
        optional['contributors'].append(
            {'contributor_type': 'DataManager', 'full_name': maintainer}
        )

    # DATES
    # created, updated, and doi publish date
    date_errors = {}
    try:
        optional['dates'].append(
            {
                'dateType': 'Created',
                'date': date_or_none(pkg_dict.get('metadata_created')),
            }
        )
    except Exception as e:
        date_errors['created'] = e
    try:
        optional['dates'].append(
            {
                'dateType': 'Updated',
                'date': date_or_none(pkg_dict.get('metadata_modified')),
            }
        )
    except Exception as e:
        date_errors['updated'] = e
    if 'doi_date_published' in pkg_dict:
        try:
            optional['dates'].append(
                {
                    'dateType': 'Issued',
                    'date': date_or_none(pkg_dict.get('doi_date_published')),
                }
            )
        except Exception as e:
            date_errors['doi_date_published'] = e

    # LANGUAGE
    # use language set in CKAN
    try:
        optional['language'] = ckan_lang()
    except Exception as e:
        errors['language'] = e

    # ALTERNATE IDENTIFIERS
    # add permalink back to this site
    try:
        permalink = f'{get_site_url()}/dataset/{pkg_dict["id"]}'
        optional['alternateIdentifiers'] = [
            {'alternateIdentifierType': 'URL', 'alternateIdentifier': permalink}
        ]
    except Exception as e:
        errors['alternateIdentifiers'] = e

    # RELATED IDENTIFIERS
    # nothing relevant in default schema

    # SIZES
    # sum up given sizes from resources in the package and convert from bytes to kilobytes
    try:
        resource_sizes = [
            r.get('size') or 0 for r in pkg_dict.get('resources', []) or []
        ]
        total_size = [f'{int(sum(resource_sizes) / 1024)} kb']
        optional['sizes'] = total_size
    except Exception as e:
        errors['sizes'] = e

    # FORMATS
    # list unique formats from package resources
    try:
        formats = list(
            set(
                filter(
                    None, [r.get('format') for r in pkg_dict.get('resources', []) or []]
                )
            )
        )
        optional['formats'] = formats
    except Exception as e:
        errors['formats'] = e

    # VERSION
    # doesn't matter if there's no version, it'll get filtered out later
    optional['version'] = pkg_dict.get('version')

    # RIGHTS
    # use the package license and get details from CKAN's license register
    license_id = pkg_dict.get('license_id', 'notspecified')
    try:
        if license_id != 'notspecified' and license_id is not None:
            license_register = Package.get_license_register()
            license = license_register.get(license_id)
            if license is not None:
                optional['rightsList'] = [
                    {'url': license.url, 'identifier': license.id}
                ]
    except Exception as e:
        errors['rightsList'] = e

    # DESCRIPTIONS
    # use package notes
    optional['descriptions'] = [
        {'descriptionType': 'Other', 'description': pkg_dict.get('notes', '')}
    ]

    # GEOLOCATIONS
    # nothing relevant in default schema

    # FUNDING
    # nothing relevant in default schema

    metadata_dict.update(required)
    metadata_dict.update(optional)

    for plugin in PluginImplementations(IDoi):
        # implementations should remove relevant errors from the errors dict if they successfully
        # handle an item
        metadata_dict, errors = plugin.build_metadata_dict(
            pkg_dict, metadata_dict, errors
        )

    for k in required:
        if metadata_dict.get(k) is None and errors.get(k) is None:
            errors[k] = DOIMetadataException('Required field cannot be None')

    required_errors = {k: e for k, e in errors.items() if k in required}
    if len(required_errors) > 0:
        error_msg = (
            f'Could not extract metadata for the following required keys: '
            f'{", ".join(required_errors)}'
        )
        log.exception(error_msg)
        for k, e in required_errors.items():
            log.exception(f'{k}: {e}')
        raise DOIMetadataException(error_msg)

    optional_errors = {k: e for k, e in errors.items() if k in optional}
    if len(required_errors) > 0:
        error_msg = (
            f'Could not extract metadata for the following optional keys: '
            f'{", ".join(optional_errors)}'
        )
        log.debug(error_msg)
        for k, e in optional_errors.items():
            log.debug(f'{k}: {e}')

    return metadata_dict


def build_xml_dict(metadata_dict):
    """
    Builds a dictionary that can be passed directly to datacite.schema42.tostring() to
    generate xml. Previously named metadata_to_xml but renamed as it's not actually
    producing any xml, it's just formatting the metadata so a separate function can then
    generate the xml.

    :param metadata_dict: a dict of metadata generated from build_metadata_dict
    :return: dict that can be passed directly to datacite.schema42.tostring()
    """

    # required fields first (DOI will be added later)
    xml_dict = {
        'creators': [],
        'titles': metadata_dict.get('titles', []),
        'publisher': metadata_dict.get('publisher'),
        'publicationYear': str(metadata_dict.get('publicationYear')),
        'types': {
            'resourceType': metadata_dict.get('resourceType'),
            'resourceTypeGeneral': 'Dataset',
        },
        'schemaVersion': 'http://datacite.org/schema/kernel-4',
    }

    for creator in metadata_dict.get('creators', []):
        xml_dict['creators'].append(xml_utils.create_contributor(**creator))

    optional = [
        'subjects',
        'contributors',
        'dates',
        'language',
        'alternateIdentifiers',
        'relatedIdentifiers',
        'sizes',
        'formats',
        'version',
        'rightsList',
        'descriptions',
        'geolocations',
        'fundingReferences',
    ]

    for k in optional:
        v = metadata_dict.get(k)
        try:
            has_value = v is not None and len(v) > 0
        except:
            has_value = False
        if not has_value:
            continue
        if k == 'contributors':
            xml_dict['contributors'] = []
            for contributor in v:
                xml_dict['contributors'].append(
                    xml_utils.create_contributor(**contributor)
                )
        elif k == 'dates':
            item = []
            for date_entry in v:
                date_entry_copy = {k: v for k, v in date_entry.items()}
                date_entry_copy['date'] = str(date_entry_copy['date'])
                item.append(date_entry_copy)
            xml_dict[k] = item
        else:
            xml_dict[k] = v

    for plugin in PluginImplementations(IDoi):
        xml_dict = plugin.build_xml_dict(metadata_dict, xml_dict)

    return xml_dict



#!/usr/bin/env python3
# encoding: utf-8
#
# This file is part of ckanext-doi
# Created by the Natural History Museum in London, UK

import xml.etree.ElementTree as ET

import pytest
from datacite.errors import DataCiteError, DataCiteNotFoundError
from unittest.mock import patch, MagicMock

from .helpers import constants
from ckanext.doi.lib.api import DataciteClient


def first_then(first, then):
    """
    Convenience generator which yields the first parameter and then yields the then
    parameter forever. Handy for MagicMock object side_effects where the first time the
    function is run first parameter should be return and then the then parameter after
    that.

    :param first: the first value to yield, once
    :param then: the parameter to yield on subsequent next calls, forever
    """
    yield first
    while True:
        yield then


@pytest.mark.ckan_config('ckanext.doi.prefix', 'testing')
class TestGenerateNewDOI:
    """
    In each of these tests we could assert the number of calls to the db and datacite
    client mocks but this feels like a bit too much of a reach into the internal logic
    of the function.

    We care about the result based on the scenario not how it gets there (unless we
    think we should test that the function checks the db before checking datacite for
    performance but I don't think this is really needed tbh.
    """

    def test_no_existing_dois(self):
        # no dois in datacite
        mock_client = MagicMock(
            metadata_get=MagicMock(side_effect=DataCiteNotFoundError())
        )
        # no dois in the database
        mock_read_doi = MagicMock(return_value=None)

        with patch(
            'ckanext.doi.lib.api.DataCiteMDSClient', MagicMock(return_value=mock_client)
        ):
            with patch('ckanext.doi.lib.api.DOIQuery.read_doi', mock_read_doi):
                api = DataciteClient()
                doi = api.generate_doi()
                assert isinstance(doi, str)
                # both the client and the database should be called once and only once (yes this
                # goes against the comment at the start of this class but it felt relevant here to
                # check that this was the case)
                assert mock_client.metadata_get.call_count == 1
                assert mock_read_doi.call_count == 1

    def test_one_existing_db_doi(self):
        # no dois in datacite
        mock_client = MagicMock(
            metadata_get=MagicMock(side_effect=DataCiteNotFoundError())
        )
        # one doi in the database that hits the first call, but then the next time is fine
        mock_read_doi = MagicMock(side_effect=first_then(MagicMock(), None))

        with patch(
            'ckanext.doi.lib.api.DataCiteMDSClient', MagicMock(return_value=mock_client)
        ):
            with patch('ckanext.doi.lib.api.DOIQuery.read_doi', mock_read_doi):
                api = DataciteClient()
                doi = api.generate_doi()
                assert isinstance(doi, str)

    def test_one_existing_on_datacite(self):
        # the first call to the datacite client returns a (mock) doi but then the next one succeeds
        mock_client = MagicMock(
            metadata_get=MagicMock(
                side_effect=first_then(MagicMock(), DataCiteNotFoundError())
            )
        )
        # no dois in the db
        mock_read_doi = MagicMock(return_value=None)

        with patch(
            'ckanext.doi.lib.api.DataCiteMDSClient', MagicMock(return_value=mock_client)
        ):
            with patch('ckanext.doi.lib.api.DOIQuery.read_doi', mock_read_doi):
                api = DataciteClient()
                doi = api.generate_doi()
                assert isinstance(doi, str)

    def test_one_existing_on_datacite_and_one_in_the_db(self):
        # the first call to the datacite client returns a (mock) doi but then the next one succeeds
        mock_client = MagicMock(
            metadata_get=MagicMock(
                side_effect=first_then(MagicMock(), DataCiteNotFoundError())
            )
        )
        # the first call to the db returns a result but then after that we're all good
        mock_read_doi = MagicMock(side_effect=first_then(MagicMock(), None))

        with patch(
            'ckanext.doi.lib.api.DataCiteMDSClient', MagicMock(return_value=mock_client)
        ):
            with patch('ckanext.doi.lib.api.DOIQuery.read_doi', mock_read_doi):
                api = DataciteClient()
                doi = api.generate_doi()
                assert isinstance(doi, str)

    def test_it_fails_when_it_cannot_generate_a_unique_doi(self):
        # the datacite client returns an existing (mock) doi every time, so unlikely!
        mock_client = MagicMock(metadata_get=MagicMock())
        # the db returns an existing (mock) doi every time, so unlikely!
        mock_read_doi = MagicMock()

        with patch('ckanext.doi.lib.api.DataCiteMDSClient', mock_client):
            with patch('ckanext.doi.lib.api.DOIQuery.read_doi', mock_read_doi):
                api = DataciteClient()
                with pytest.raises(Exception, match='Failed to generate a DOI'):
                    api.generate_doi()


class MockDataciteMDSClient(object):
    """
    Mock client so that we can replicate the functionality of the datacite API without
    actually calling it.

    Specifically, you have to post the metadata before you post the doi.
    """

    def __init__(self, *args, **kwargs):
        self.metadata = set()
        self.dois = set()

    def doi_post(self, doi, *args, **kwargs):
        if doi not in self.metadata:
            raise DataCiteError()

        self.dois.add(doi)

    def metadata_post(self, xml_doc, *args, **kwargs):
        tree = ET.fromstring(xml_doc)
        doi = tree.findtext('{http://datacite.org/schema/kernel-4}identifier')
        self.metadata.add(doi)


@pytest.mark.ckan_config('ckanext.doi.prefix', 'testing')
@patch('ckanext.doi.lib.api.DataCiteMDSClient', MockDataciteMDSClient)
@patch('ckanext.doi.lib.api.DOIQuery')
class TestMintNewDOI(object):
    def test_datacite_api_order(self, mock_crud):
        mock_crud.read_doi = MagicMock(return_value=None)
        mock_crud.read_package = MagicMock(return_value=None)

        api = DataciteClient()
        doi = constants.XML_DICT['identifiers'][0]['identifier']
        pkg_id = MagicMock()

        with pytest.raises(DataCiteError):
            api.mint_doi(doi, pkg_id)

        api.set_metadata(doi, constants.XML_DICT)
        api.mint_doi(doi, pkg_id)

    def test_new_doi(self, mock_crud):
        mock_crud.read_doi = MagicMock(return_value=None)
        mock_crud.read_package = MagicMock(return_value=None)

        api = DataciteClient()
        doi = constants.XML_DICT['identifiers'][0]['identifier']
        pkg_id = MagicMock()

        api.set_metadata(doi, constants.XML_DICT)
        api.mint_doi(doi, pkg_id)

        assert mock_crud.create.called
        assert not mock_crud.update_package.called
        assert mock_crud.update_doi.called

    def test_existing_doi(self, mock_crud):
        mock_crud.read_doi = MagicMock(return_value=MagicMock())
        mock_crud.read_package = MagicMock(return_value=None)

        api = DataciteClient()
        doi = constants.XML_DICT['identifiers'][0]['identifier']
        pkg_id = MagicMock()

        api.set_metadata(doi, constants.XML_DICT)
        api.mint_doi(doi, pkg_id)

        assert not mock_crud.create.called
        assert not mock_crud.update_package.called
        assert mock_crud.update_doi.called

    def test_existing_package(self, mock_crud):
        mock_crud.read_doi = MagicMock(return_value=None)
        mock_crud.read_package = MagicMock(return_value=MagicMock())

        api = DataciteClient()
        doi = constants.XML_DICT['identifiers'][0]['identifier']
        pkg_id = MagicMock()

        api.set_metadata(doi, constants.XML_DICT)
        api.mint_doi(doi, pkg_id)

        assert not mock_crud.create.called
        assert mock_crud.update_package.called
        assert mock_crud.update_doi.called

    def test_both_exist(self, mock_crud):
        mock_crud.read_doi = MagicMock(return_value=MagicMock())
        mock_crud.read_package = MagicMock(return_value=MagicMock())

        api = DataciteClient()
        doi = constants.XML_DICT['identifiers'][0]['identifier']
        pkg_id = MagicMock()

        api.set_metadata(doi, constants.XML_DICT)
        api.mint_doi(doi, pkg_id)

        assert not mock_crud.create.called
        assert not mock_crud.update_package.called
        assert mock_crud.update_doi.called


@pytest.mark.ckan_config('ckanext.doi.prefix', 'testing')
@pytest.mark.ckan_config('ckanext.doi.account_name', 'goat!')
@pytest.mark.ckan_config('ckanext.doi.account_password', 'hammocks?')
@patch('ckanext.doi.lib.api.DataCiteMDSClient')
class TestDataciteClientCreation(object):
    @pytest.mark.ckan_config('ckanext.doi.test_mode', False)
    def test_basics(self, mock_client):
        DataciteClient()
        assert mock_client.called_once_with(
            username='goat!', password='hammocks?', prefix='testing', test_mode=False
        )
        assert 'url' not in mock_client.call_args.kwargs

    @pytest.mark.ckan_config('ckanext.doi.test_mode', True)
    def test_test_mode_true(self, mock_client):
        DataciteClient()
        assert mock_client.called_once_with(
            username='goat!',
            password='hammocks?',
            prefix='testing',
            test_mode=True,
            url=DataciteClient.test_url,
        )



#!/usr/bin/env python3
# encoding: utf-8
#
# This file is part of ckanext-doi
# Created by the Natural History Museum in London, UK

from ckan.model import Session

from ckanext.doi.model.doi import DOI, doi_table


class DOIQuery:
    # convenience properties
    m = DOI
    cols = [c.name for c in doi_table.c]

    @classmethod
    def create(cls, identifier, package_id, published=None):
        """
        Create a new record in the DOI table.

        :param identifier: a new DOI string
        :param package_id: the id of the package this DOI represents
        :param published: when this DOI was published (datetime, nullable)
        :return: the newly created record object
        """
        new_record = DOI(
            identifier=identifier, package_id=package_id, published=published
        )
        Session.add(new_record)
        Session.commit()
        return new_record

    @classmethod
    def read_doi(cls, identifier):
        """
        Retrieve a record with a given DOI.

        :param identifier: the DOI string
        :return: the record object
        """
        return Session.query(DOI).get(identifier)

    @classmethod
    def read_package(cls, package_id, create_if_none=False):
        """
        Retrieve a record associated with a given package.

        :param package_id: the id of the package
        :param create_if_none: generate a new DOI and add a record if no record is found for the
                               given package
        :return: the record object
        """
        from ckanext.doi.lib.api import DataciteClient

        record = Session.query(DOI).filter(DOI.package_id == package_id).first()
        if record is None and create_if_none:
            client = DataciteClient()
            new_doi = client.generate_doi()
            record = cls.create(new_doi, package_id)
        return record

    @classmethod
    def update_doi(cls, identifier, **kwargs):
        """
        Update the package_id and/or published fields of a record with a given DOI.

        :param identifier: the DOI string
        :param kwargs: the values to be updated
        :return: the updated record object
        """
        update_dict = {k: v for k, v in kwargs.items() if k in cls.cols}
        Session.query(DOI).filter(DOI.identifier == identifier).update(update_dict)
        Session.commit()
        return cls.read_doi(identifier)

    @classmethod
    def update_package(cls, package_id, **kwargs):
        """
        Update the package_id and/or published fields of a record associated with a
        given package.

        :param package_id: the id of the package
        :param kwargs: the values to be updated
        :return: the updated record object
        """
        update_dict = {k: v for k, v in kwargs.items() if k in cls.cols}
        Session.query(DOI).filter(DOI.package_id == package_id).update(update_dict)
        Session.commit()
        return cls.read_package(package_id)

    @classmethod
    def delete_doi(cls, identifier):
        """
        Delete the record with a given DOI.

        :param identifier: the DOI string
        :return: True if a record was deleted, False if not
        """
        to_delete = cls.read_doi(identifier)
        if to_delete is not None:
            Session.delete(to_delete)
            Session.commit()
            return True
        else:
            return False

    @classmethod
    def delete_package(cls, package_id):
        """
        Delete the record associated with a given package.

        :param package_id: the id of the package
        :return: True if a record was deleted, False if not
        """
        to_delete = cls.read_package(package_id)
        if to_delete is not None:
            Session.delete(to_delete)
            Session.commit()
            return True
        else:
            return False



import click
from ckan import model
from ckan.model import Session
from ckan.plugins import toolkit
from datacite.errors import DataCiteError

from ckanext.doi.lib.api import DataciteClient
from ckanext.doi.lib.metadata import build_metadata_dict, build_xml_dict
from ckanext.doi.model import doi as doi_model
from ckanext.doi.model.crud import DOIQuery
from ckanext.doi.model.doi import DOI


def get_commands():
    return [doi]


@click.group()
def doi():
    pass


@doi.command(name='initdb')
def init_db():
    if not model.package_table.exists():
        click.secho(
            'Package table must exist before initialising the DOI table', fg='red'
        )
        raise click.Abort()

    if doi_model.doi_table.exists():
        click.secho('DOI table already exists', fg='green')
    else:
        doi_model.doi_table.create()
        click.secho('DOI table created', fg='green')


@doi.command(name='delete-dois')
def delete_dois():
    """
    Delete all DOIs from the database.
    """
    to_delete = Session.query(DOI).filter(
        DOI.identifier.like(f'%{DataciteClient.get_prefix()}%')
    )
    doi_count = to_delete.count()
    if doi_count == 0:
        click.secho('Nothing to delete', fg='green')
        return
    if click.confirm(f'Delete {doi_count} DOIs from the database?', abort=True):
        to_delete.delete(synchronize_session=False)
        Session.commit()
        click.secho(f'Deleted {doi_count} DOIs from the database')


@doi.command(name='update-doi')
@click.option(
    '-p', '--package_id', 'package_ids', multiple=True, help='Package id(s) to update'
)
def update_doi(package_ids):
    """
    Update either all DOIs in the system or the ones associated with the given packages.
    """
    if not package_ids:
        dois_to_update = Session.query(DOI).all()
    else:
        dois_to_update = list(filter(None, map(DOIQuery.read_package, package_ids)))

    if len(dois_to_update) == 0:
        click.secho('No DOIs found to update', fg='green')
        return

    for record in dois_to_update:
        pkg_dict = toolkit.get_action('package_show')({}, {'id': record.package_id})
        title = pkg_dict.get('title', record.package_id)

        if record.published is None:
            click.secho(
                f'"{title}" does not have a published DOI; ignoring', fg='yellow'
            )
            continue
        if pkg_dict.get('state', 'active') != 'active' or pkg_dict.get(
            'private', False
        ):
            click.secho(f'"{title}" is inactive or private; ignoring', fg='yellow')
            continue

        metadata_dict = build_metadata_dict(pkg_dict)
        xml_dict = build_xml_dict(metadata_dict)

        client = DataciteClient()

        same = client.check_for_update(record.identifier, xml_dict)
        if not same:
            try:
                client.set_metadata(record.identifier, xml_dict)
                click.secho(f'Updated "{title}"', fg='green')
            except DataCiteError as e:
                click.secho(
                    f'Error while updating "{title}" (DOI {record.identifier}): {e})',
                    fg='red',
                )
        else:
            click.secho(f'"{title}" is already up to date', fg='green')



#!/usr/bin/env python3
# encoding: utf-8
#
# This file is part of ckanext-doi
# Created by the Natural History Museum in London, UK
import pytest
from ckan.model import Session
from ckan.tests import factories
from datacite.errors import DataCiteNotFoundError
from unittest.mock import patch, MagicMock

from ckanext.doi.model.doi import DOI, doi_table


@pytest.fixture
def with_doi_table(reset_db):
    """
    Simple fixture which resets the database and creates the doi table.
    """
    reset_db()
    doi_table.create(checkfirst=True)


@pytest.mark.filterwarnings('ignore::sqlalchemy.exc.SADeprecationWarning')
@pytest.mark.ckan_config('ckan.plugins', 'doi')
@pytest.mark.ckan_config('ckanext.doi.prefix', 'testing')
@pytest.mark.usefixtures('with_doi_table', 'with_plugins')
@patch('ckanext.doi.lib.api.DataCiteMDSClient')
def test_doi_is_created_automatically(mock_client):
    # this test ensures that when a package is created, a DOI is created
    mock_client.return_value = MagicMock(
        metadata_get=MagicMock(side_effect=DataCiteNotFoundError())
    )
    package = factories.Dataset()
    found_record = Session.query(DOI).filter(DOI.package_id == package['id']).one()
    assert found_record is not None
